

#########################################################
#                                                       #
#     MOUSE ARTIFICIAL IMMUNE SYSTEM ALGORITHMS         #
#                   (02.09.2019)                        #
#                 R. Lahoz-Beltra                       #
#                                                       #    
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND   #
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY #
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  #
# THE SOFWTARE CAN BE USED BY ANYONE SOLELY FOR THE     #
# PURPOSES OF RESEARCH AND EDUCATION.                   #
#                                                       #
#########################################################
import math
import random
from random import randint
import numpy as np
import re
RNAtarget= np.empty([100],dtype=np.int) 

#########################################################
# ALGORITHM PARAMETERS                                  #
#########################################################
N= 80                # Population size
Genome= 64           # Genome length 
popSize = N+1
genomeLength  = Genome+1

#########################################################
# HOST ALGORITHM PARAMETERS                             #
#########################################################
paratope_length=16;
# B-cells population size
bcell = np.empty([popSize,paratope_length],dtype=np.int)
nbcell = np.empty([popSize,paratope_length],dtype=np.int)

# dendritic-cells 
organs=12
DCell = np.empty([organs])
PS = np.empty([organs])
DS = np.empty([organs])
SS = np.empty([organs])
IC = np.empty([organs])

# Signals_Stop_Infection parameters
stop = np.empty([4],dtype=np.int)

# variables 
fitness_bcell = np.empty([popSize])   
probability_ = np.empty([popSize])                          
net_ = np.empty([popSize])

maplength = 24
host = np.empty([organs,maplength],dtype=np.int)               
virus= np.empty([maplength],dtype=np.int) 

#########################################################
# RANDOM NUMBER GENERATOR                               #
#########################################################
def qseed():
    np.random.seed()

#########################################################
# TARGET VIRUS                                          #
#########################################################
def RNA(rnafilename):
    fh = open(rnafilename,'r')
    file = fh.read()
    nucleic = re.findall('[01]',file)
    RNA_ = np.array(nucleic,dtype=int)
    # print('RNA: ',RNA)
    fh.close()
    for i in range(0,genomeLength-1):
       RNAtarget[i]=RNA_[i]
       print(RNAtarget[i], end= " ")
    print()
 
#########################################################
# RAT HOST DEFINITION                                   #
######################################################### 
def Host(hostfile):  
    fh = open(hostfile,'r')
    file = fh.read()
    rat = re.findall('[01]',file)
    cont = 0  
    for i in range(0,organs):
        for j in range(0,maplength):
            host[i,j]=rat[cont]
            cont +=1
    fh.close()

#########################################################
# NEGATIVE SELECTION ALGORITHM                          #
#########################################################         
def Negative_selection(filename):
    f = open(filename,"a")
    cells=1
    
    H=np.empty([organs],dtype=np.int)
    H_=paratope_length*organs
    detection_thr=24
    while (H_>detection_thr):
        H_=0
        for i in range (0, organs):
            H[i]=0
                    
        print("B-cell      ",end=" ")    
        # Init_Bcell()
        for j in range(0,paratope_length):  
            u=np.random.randint(0,100)
            u=u/100
            if u<=0.5:
                bcell[cells,j]=0
            else:
                bcell[cells,j]=1
            print(bcell[cells,j],end=" ")
        print()
           
        for j in range (0, organs):
            for i in range (0, paratope_length):
                if bcell[cells,i] != host[j,i+8]:  
                    H[j] +=1
        for i in range (0, organs):
            H_ = H_ + H[i]            
        print("  H = ",H_)
        print()
        if H_<=detection_thr:
            break
    
    print()  
    print("B-cell      ",end=" ")    
    for i in range(0,paratope_length):
         print(bcell[cells,i],end=" ")
         f.write(str(bcell[cells,i]))
        
    print()
    f.write("\n")
    f.close()

#########################################################
# B-CELLS MATURE POPULATION INITIALIZATION              #
#########################################################
def Set_Tcell(filename, N):
    i=0
    while i<N:
        Negative_selection(filename)
        i+=1
        if i==N:
            break    
        
#########################################################
# B-CELLS INMATURE INITIALIZATION                       #
#########################################################
def Init_bcellspopulation():
    print("Random population of B-cells: ",end="")
    print()
    print()
    for i in range(1,popSize):
        for j in range(0,paratope_length):  
            u=np.random.randint(0,100)
            u=u/100
            if u<=0.5:
                bcell[i,j]=0
            else:
                bcell[i,j]=1
            print(bcell[i,j],end=" ")
        print()
              
def Show_bcells():
    print()
    print("Population of B-cells: ",end="")
    print()
    print()
    for i in range(1,popSize):
        for j in range(0,16):
            print(bcell[i,j],end="")
        print()  

#########################################################
# HUMORAL SYSTEM ALGORITHM                              #
#########################################################
def nonself_self(rule, threshold,stop_generation,generation):
    stop[0]=0;
    b=np.empty([N+1])
    for i in range(1,N+1):
        b[i]=fitness_bcell[i]
        bcell_selected=b.argmax()
    print()
    print("bcell selected is:  ",bcell_selected," ",fitness_bcell[bcell_selected]," : \n")
    for j in range(0,16):
        print(bcell[bcell_selected,j],end="")
    print()  
   
    # stop(generation_s, bcell_selected)
    H_=0
    for j in range(0, paratope_length):
        if bcell[bcell_selected,j] != virus[j+8]:
            H_+=1
    if rule==0:
        if H_>=threshold and generation==stop_generation:
            stop[0]=1;
    if rule==1:
        if H_>=threshold or generation==stop_generation:
            stop[0]=1;
    print()
    print("state of infection : ",stop[0],"  \n")
    
#########################################################
# DENDRITIC CELL ALGORITHM                              #
#########################################################
def nonself_danger(tissue, SP, SD, SS, IC, threshold): 
    """k: dendritic cell integer id 
       csm: co-stimulary molecules --> CSM citokines
       semi: semimature cytokines 
       mat:  mature signals """
    
    # Intialize
    max_iterations=200
    Ocsm = np.empty([max_iterations])
    Osemi = np.empty([max_iterations])
    Omat = np.empty([max_iterations])
    
    for i in  range(1,max_iterations):
        Ocsm[i]=0; Osemi[i]=0; Omat[i]=0; 
  
    # Damaged organ
    organ=int(tissue[0])*8+int(tissue[1])*4+int(tissue[2])*2+int(tissue[3])
    DCell[organ]=0
    print("organ dec = ",organ,"(",tissue,")"," state = ",DCell[organ])
    
    # DCA algorithm    
    for i in range(0,100):
        if Ocsm[i]<threshold:
            Ocsm[i]=Ocsm[i-1]+(SP*2+SD+SS*2)*(1+IC)
            Osemi[i]=Osemi[i-1]+(SS*3)*(1+IC)
            Omat[i]=Omat[i-1]+(SP*2+SD-SS*3)*(1+IC)
            print("t = ",i," Ocsm = ",Ocsm[i]," Osemi = ",Osemi[i]," Omat = ",Omat[i],end=" ")
            print()
        if Ocsm[i]>threshold:
            if Osemi[i]>Omat[i]:
                print("no danger, suppresses T-cell ",DCell[organ])
            if Osemi[i]<Omat[i]:
                DCell[organ]=1  
                print("danger, stimulates T-cell ",DCell[organ])          
            break

#########################################################
#   CLONAL SELECTION ALGORITHM                          #
#########################################################  
def AgAb_recognition(generation_s, filename):
    fitness_total_bcell = 0; fitness_average_bcell = 0; 
    for i in range(1,popSize):
        H=0
        fitness_bcell[i]=0
        for j in range (0, paratope_length):
            if bcell[i,j] != virus[j+8]:
                H+=1
        # Fitness function:        
        fitness_bcell[i]=H*10    
        print()  
        print("bcell ",i," fitness = ",fitness_bcell[i],"Hamming = ",H)
           
    for i in range(1,N+1):
        fitness_total_bcell = fitness_total_bcell + fitness_bcell[i]
    fitness_average_bcell = fitness_total_bcell/N
    print('fitness average = ',fitness_average_bcell)
    f = open(filename, "a")
    f.write(str(generation_s)+" "+str(fitness_average_bcell)+"\n")
    f.write(" \n")
    f.close()
    
def Clonal_selection():
    fitness_total_bcell=0; parent=0;
    
    for i in range(1,popSize):
        fitness_total_bcell = fitness_total_bcell + fitness_bcell[i]
    for i in range(1,popSize):
        probability_[i]=fitness_bcell[i]/fitness_total_bcell
    for i in range(1,popSize):
        net_[0]=0
    for i in range(1,popSize):
        net_[i]=net_[i-1]+probability_[i]

    u=np.random.uniform(0,1)
    bugs=1; k=1;
    while bugs<=N:
        for i in range(1,popSize):
            if u<net_[i]:
                parent=i
                for j in range(0,paratope_length):
                    nbcell[k,j]=bcell[parent,j]
                k=k+1
                break
           
        u=np.random.uniform(0,1)   
        bugs=bugs+1
    for i in range(1,popSize):
        for j in range(0,paratope_length):
            bcell[i,j]=nbcell[i,j]
            
def hypermutation(mutation_rate, delta):
    u=0;
    for i in range(1,popSize):
        up=np.random.random_integers(100)
        up=up/100
        # Probar otras funciones:
        if up<=math.exp(-fitness_bcell[i]/delta):
            for j in range(0,paratope_length):
                u=np.random.random_integers(100)
                u=u/100
                if u<=mutation_rate:
                    if bcell[i,j]==0:
                        nbcell[i,j]=1
                    else:
                        nbcell[i,j]=0
                else:
                    nbcell[i,j]=bcell[i,j]
        else:
            for j in range(0,paratope_length):
                nbcell[i,j]=bcell[i,j]
    for i in range(1,popSize):
        for j in range(0,paratope_length):
            bcell[i,j]=nbcell[i,j]

def sCLONAL(filename1):
    qseed()
    generation_s=0
    generation_max_s=500
    print("============== GENERATION: ",generation_s," =========================== ")
    print()
    Init_bcellspopulation()
    RNA(filename1)
    AgAb_recognition(0,'B-cells.dat')
    while (generation_s<generation_max_s-1):
        print()
        print("============== GENERATION: ",generation_s+1," =========================== ")
        print()
        # Select B-cells, update generation number and obtain fitness
        Clonal_selection()
        generation_s=generation_s+1
        AgAb_recognition(generation_s,'B-cells.dat')
        # Apply hipermutation operator
        hypermutation(0.8,0.5)
    Show_bcells()
 
#########################################################
# HUMORAL DETECTION ALGORITHM                           #
#########################################################
def Detection(filename1, filename2):
    RNA(filename1)
    Host(filename2) 
    stop[1]=0
    u=np.random.randint(0,12)
    H=0
    for i in range (0, 10):
        if host[u,i+8] != virus[i+8]:
            H+=1
    if H!=0:
        stop[1]=1
        sCLONAL(filename1)
    print("foreign agent detected : ",stop[1]) 

#########################################################
# DANGER DETECTION ALGORITHM                            #
#########################################################
def dDetection(filename1, filename2, tissue, threshold, stop_generation, generation):
    if generation==1:
        Detection(filename1, filename2)
                
        organ=int(tissue[0])*8+int(tissue[1])*4+int(tissue[2])*2+int(tissue[3])
        DCell[organ]=0
        PS[organ]=0; DS[organ]=0; SS[organ]=0; IC[organ]=0;
        
    else:
        organ=int(tissue[0])*8+int(tissue[1])*4+int(tissue[2])*2+int(tissue[3])
        # Signals model 0 = 0.01 / 1 = 0.1
        signal_threshold=10
        if stop[1]==1:
            PS[organ]+=0.001
            # DS[organ]+=0.01
            # SS[organ]+=0.01
            if PS[organ] > 0.5 or DS[organ] > 0.5:
                IC[organ]=1.0
        
        print("Input signals :  PAMP = ",PS[organ]," Danger = ",DS[organ]," Safe = ",SS[organ],end=" ")
        print("Inflamation = ",IC[organ])
        nonself_danger(tissue,PS[organ],DS[organ],SS[organ],IC[organ], signal_threshold)   
        if DCell[organ]==1:
            nonself_self(1, threshold,stop_generation,generation)
    
